// Copyright 2020, 2021, 2022 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
//
// This file is part of the heyoka library.
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef HEYOKA_MATH_HPP
#define HEYOKA_MATH_HPP

#include <heyoka/math/acos.hpp>
#include <heyoka/math/acosh.hpp>
#include <heyoka/math/asin.hpp>
#include <heyoka/math/asinh.hpp>
#include <heyoka/math/atan.hpp>
#include <heyoka/math/atan2.hpp>
#include <heyoka/math/atanh.hpp>
#include <heyoka/math/binary_op.hpp>
#include <heyoka/math/constants.hpp>
#include <heyoka/math/cos.hpp>
#include <heyoka/math/cosh.hpp>
#include <heyoka/math/erf.hpp>
#include <heyoka/math/exp.hpp>
#include <heyoka/math/kepE.hpp>
#include <heyoka/math/log.hpp>
#include <heyoka/math/neg.hpp>
#include <heyoka/math/pow.hpp>
#include <heyoka/math/sigmoid.hpp>
#include <heyoka/math/sin.hpp>
#include <heyoka/math/sinh.hpp>
#include <heyoka/math/sqrt.hpp>
#include <heyoka/math/square.hpp>
#include <heyoka/math/sum.hpp>
#include <heyoka/math/sum_sq.hpp>
#include <heyoka/math/tan.hpp>
#include <heyoka/math/tanh.hpp>
#include <heyoka/math/time.hpp>
#include <heyoka/math/tpoly.hpp>

#endif
